#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from openstackclient.tests.functional import base


class HypervisorTests(base.TestCase):
    """Functional tests for hypervisor."""

    def test_hypervisor_list(self):
        """Test create defaults, list filters, delete"""
        # Test list
        cmd_output = json.loads(
            self.openstack(
                "hypervisor list -f json --os-compute-api-version 2.1"
            )
        )
        ids1 = [x["ID"] for x in cmd_output]
        self.assertIsNotNone(cmd_output)

        cmd_output = json.loads(self.openstack("hypervisor list -f json"))
        ids2 = [x["ID"] for x in cmd_output]
        self.assertIsNotNone(cmd_output)

        # Show test - old microversion
        for i in ids1:
            cmd_output = json.loads(
                self.openstack(
                    f"hypervisor show {i} -f json "
                    " --os-compute-api-version 2.1"
                )
            )
            self.assertIsNotNone(cmd_output)
        # When we list hypervisors with older MV we get ids as integers. We
        # need to verify that show finds resources independently
        # Show test - latest microversion
        for i in ids2:
            cmd_output = json.loads(
                self.openstack(f"hypervisor show {i} -f json")
            )
            self.assertIsNotNone(cmd_output)



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from osc_lib import exceptions

from openstackclient.network.v2 import network_flavor
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestNetworkFlavor(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains


class TestAddNetworkFlavorToProfile(TestNetworkFlavor):
    network_flavor = network_fakes.create_one_network_flavor()
    service_profile = network_fakes.create_one_service_profile()

    def setUp(self):
        super().setUp()

        self.network_client.find_flavor.return_value = self.network_flavor
        self.network_client.find_service_profile.return_value = (
            self.service_profile
        )

        self.cmd = network_flavor.AddNetworkFlavorToProfile(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_add_flavor_to_service_profile(self):
        arglist = [self.network_flavor.id, self.service_profile.id]
        verifylist = [
            ('flavor', self.network_flavor.id),
            ('service_profile', self.service_profile.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.network_client.associate_flavor_with_service_profile.assert_called_once_with(  # noqa: E501
            self.network_flavor, self.service_profile
        )


class TestCreateNetworkFlavor(TestNetworkFlavor):
    project = identity_fakes_v3.FakeProject.create_one_project()
    domain = identity_fakes_v3.FakeDomain.create_one_domain()
    # The new network flavor created.
    new_network_flavor = network_fakes.create_one_network_flavor()
    columns = (
        'description',
        'enabled',
        'id',
        'name',
        'service_type',
        'service_profile_ids',
    )
    data = (
        new_network_flavor.description,
        new_network_flavor.is_enabled,
        new_network_flavor.id,
        new_network_flavor.name,
        new_network_flavor.service_type,
        new_network_flavor.service_profile_ids,
    )

    def setUp(self):
        super().setUp()
        self.network_client.create_flavor.return_value = (
            self.new_network_flavor
        )

        # Get the command object to test
        self.cmd = network_flavor.CreateNetworkFlavor(self.app, None)

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            '--service-type',
            self.new_network_flavor.service_type,
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('service_type', self.new_network_flavor.service_type),
            ('name', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_flavor.assert_called_once_with(
            **{
                'service_type': self.new_network_flavor.service_type,
                'name': self.new_network_flavor.name,
            }
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))

    def test_create_all_options(self):
        arglist = [
            '--description',
            self.new_network_flavor.description,
            '--enable',
            '--project',
            self.project.id,
            '--project-domain',
            self.domain.name,
            '--service-type',
            self.new_network_flavor.service_type,
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('description', self.new_network_flavor.description),
            ('enable', True),
            ('project', self.project.id),
            ('project_domain', self.domain.name),
            ('service_type', self.new_network_flavor.service_type),
            ('name', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_flavor.assert_called_once_with(
            **{
                'description': self.new_network_flavor.description,
                'enabled': True,
                'project_id': self.project.id,
                'service_type': self.new_network_flavor.service_type,
                'name': self.new_network_flavor.name,
            }
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))

    def test_create_disable(self):
        arglist = [
            '--disable',
            '--service-type',
            self.new_network_flavor.service_type,
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('disable', True),
            ('service_type', self.new_network_flavor.service_type),
            ('name', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_flavor.assert_called_once_with(
            **{
                'enabled': False,
                'service_type': self.new_network_flavor.service_type,
                'name': self.new_network_flavor.name,
            }
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))


class TestDeleteNetworkFlavor(TestNetworkFlavor):
    # The network flavor to delete.
    _network_flavors = network_fakes.create_flavor(count=2)

    def setUp(self):
        super().setUp()
        self.network_client.delete_flavor.return_value = None
        self.network_client.find_flavor = network_fakes.get_flavor(
            network_flavors=self._network_flavors
        )

        # Get the command object to test
        self.cmd = network_flavor.DeleteNetworkFlavor(self.app, None)

    def test_network_flavor_delete(self):
        arglist = [
            self._network_flavors[0].name,
        ]
        verifylist = [
            ('flavor', [self._network_flavors[0].name]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network_client.find_flavor.assert_called_once_with(
            self._network_flavors[0].name, ignore_missing=False
        )
        self.network_client.delete_flavor.assert_called_once_with(
            self._network_flavors[0]
        )
        self.assertIsNone(result)

    def test_multi_network_flavors_delete(self):
        arglist = []
        verifylist = []

        for a in self._network_flavors:
            arglist.append(a.name)
        verifylist = [
            ('flavor', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for a in self._network_flavors:
            calls.append(mock.call(a))
        self.network_client.delete_flavor.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_network_flavors_delete_with_exception(self):
        arglist = [
            self._network_flavors[0].name,
            'unexist_network_flavor',
        ]
        verifylist = [
            (
                'flavor',
                [self._network_flavors[0].name, 'unexist_network_flavor'],
            ),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._network_flavors[0], exceptions.CommandError]
        self.network_client.find_flavor.side_effect = find_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 flavors failed to delete.', str(e))

        self.network_client.find_flavor.assert_any_call(
            self._network_flavors[0].name, ignore_missing=False
        )
        self.network_client.find_flavor.assert_any_call(
            'unexist_network_flavor', ignore_missing=False
        )
        self.network_client.delete_flavor.assert_called_once_with(
            self._network_flavors[0]
        )


class TestListNetworkFlavor(TestNetworkFlavor):
    # The network flavors to list up.
    _network_flavors = network_fakes.create_flavor(count=2)
    columns = (
        'ID',
        'Name',
        'Enabled',
        'Service Type',
        'Description',
    )
    data = []
    for flavor in _network_flavors:
        data.append(
            (
                flavor.id,
                flavor.name,
                flavor.is_enabled,
                flavor.service_type,
                flavor.description,
            )
        )

    def setUp(self):
        super().setUp()
        self.network_client.flavors.return_value = self._network_flavors

        # Get the command object to test
        self.cmd = network_flavor.ListNetworkFlavor(self.app, None)

    def test_network_flavor_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.flavors.assert_called_once_with(**{})
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(self.data, list(data))


class TestRemoveNetworkFlavorFromProfile(TestNetworkFlavor):
    network_flavor = network_fakes.create_one_network_flavor()
    service_profile = network_fakes.create_one_service_profile()

    def setUp(self):
        super().setUp()
        self.network_client.find_flavor.return_value = self.network_flavor
        self.network_client.find_service_profile.return_value = (
            self.service_profile
        )
        self.network_client.disassociate_flavor_from_service_profile = (
            mock.Mock()
        )

        self.cmd = network_flavor.RemoveNetworkFlavorFromProfile(
            self.app, None
        )

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_remove_flavor_from_service_profile(self):
        arglist = [self.network_flavor.id, self.service_profile.id]
        verifylist = [
            ('flavor', self.network_flavor.id),
            ('service_profile', self.service_profile.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.network_client.disassociate_flavor_from_service_profile.assert_called_once_with(  # noqa: E501
            self.network_flavor, self.service_profile
        )


class TestShowNetworkFlavor(TestNetworkFlavor):
    # The network flavor to show.
    new_network_flavor = network_fakes.create_one_network_flavor()
    columns = (
        'description',
        'enabled',
        'id',
        'name',
        'service_type',
        'service_profile_ids',
    )
    data = (
        new_network_flavor.description,
        new_network_flavor.is_enabled,
        new_network_flavor.id,
        new_network_flavor.name,
        new_network_flavor.service_type,
        new_network_flavor.service_profile_ids,
    )

    def setUp(self):
        super().setUp()
        self.network_client.find_flavor.return_value = self.new_network_flavor

        # Get the command object to test
        self.cmd = network_flavor.ShowNetworkFlavor(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Missing required args should bail here
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('flavor', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_flavor.assert_called_once_with(
            self.new_network_flavor.name, ignore_missing=False
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertEqual(set(self.data), set(data))


class TestSetNetworkFlavor(TestNetworkFlavor):
    # The network flavor to set.
    new_network_flavor = network_fakes.create_one_network_flavor()

    def setUp(self):
        super().setUp()
        self.network_client.update_flavor.return_value = None
        self.network_client.find_flavor.return_value = self.new_network_flavor

        # Get the command object to test
        self.cmd = network_flavor.SetNetworkFlavor(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('flavor', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {}
        self.network_client.update_flavor.assert_called_with(
            self.new_network_flavor, **attrs
        )
        self.assertIsNone(result)

    def test_set_name_and_enable(self):
        arglist = [
            '--name',
            'new_network_flavor',
            '--enable',
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('name', 'new_network_flavor'),
            ('enable', True),
            ('flavor', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'name': "new_network_flavor",
            'enabled': True,
        }
        self.network_client.update_flavor.assert_called_with(
            self.new_network_flavor, **attrs
        )
        self.assertIsNone(result)

    def test_set_disable(self):
        arglist = [
            '--disable',
            self.new_network_flavor.name,
        ]
        verifylist = [
            ('disable', True),
            ('flavor', self.new_network_flavor.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'enabled': False,
        }
        self.network_client.update_flavor.assert_called_with(
            self.new_network_flavor, **attrs
        )
        self.assertIsNone(result)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Policy action implementations"""

import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class CreatePolicy(command.ShowOne):
    _description = _("Create new policy")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--type',
            metavar='<type>',
            default="application/json",
            help=_(
                'New MIME type of the policy rules file '
                '(defaults to application/json)'
            ),
        )
        parser.add_argument(
            'rules',
            metavar='<filename>',
            help=_('New serialized policy rules file'),
        )
        return parser

    def take_action(self, parsed_args):
        blob = utils.read_blob_file_contents(parsed_args.rules)

        identity_client = self.app.client_manager.identity
        policy = identity_client.policies.create(
            blob=blob, type=parsed_args.type
        )

        policy._info.pop('links')
        policy._info.update({'rules': policy._info.pop('blob')})
        return zip(*sorted(policy._info.items()))


class DeletePolicy(command.Command):
    _description = _("Delete policy(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar='<policy>',
            nargs='+',
            help=_('Policy(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        result = 0
        for i in parsed_args.policy:
            try:
                identity_client.policies.delete(i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete policy with name or "
                        "ID '%(policy)s': %(e)s"
                    ),
                    {'policy': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.policy)
            msg = _("%(result)s of %(total)s policies failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListPolicy(command.Lister):
    _description = _("List policies")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def take_action(self, parsed_args):
        columns: tuple[str, ...] = ('ID', 'Type')
        column_headers: tuple[str, ...] = columns
        if parsed_args.long:
            columns += ('Blob',)
            column_headers += ('Rules',)
        data = self.app.client_manager.identity.policies.list()
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetPolicy(command.Command):
    _description = _("Set policy properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar='<policy>',
            help=_('Policy to modify'),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            help=_('New MIME type of the policy rules file'),
        )
        parser.add_argument(
            '--rules',
            metavar='<filename>',
            help=_('New serialized policy rules file'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        blob = None

        if parsed_args.rules:
            blob = utils.read_blob_file_contents(parsed_args.rules)

        kwargs = {}
        if blob:
            kwargs['blob'] = blob
        if parsed_args.type:
            kwargs['type'] = parsed_args.type

        identity_client.policies.update(parsed_args.policy, **kwargs)


class ShowPolicy(command.ShowOne):
    _description = _("Display policy details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar='<policy>',
            help=_('Policy to display'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        policy = utils.find_resource(
            identity_client.policies, parsed_args.policy
        )

        policy._info.pop('links')
        policy._info.update({'rules': policy._info.pop('blob')})
        return zip(*sorted(policy._info.items()))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Role action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


def _format_role(role):
    columns = (
        "id",
        "name",
        "domain_id",
        "description",
    )
    column_headers = (
        "id",
        "name",
        "domain_id",
        "description",
    )
    return (
        column_headers,
        utils.get_item_properties(role, columns),
    )


def _add_identity_and_resource_options_to_parser(parser):
    system_or_domain_or_project = parser.add_mutually_exclusive_group()
    system_or_domain_or_project.add_argument(
        '--system',
        metavar='<system>',
        help=_('Include <system> (all)'),
    )
    system_or_domain_or_project.add_argument(
        '--domain',
        metavar='<domain>',
        help=_('Include <domain> (name or ID)'),
    )
    system_or_domain_or_project.add_argument(
        '--project',
        metavar='<project>',
        help=_('Include <project> (name or ID)'),
    )
    user_or_group = parser.add_mutually_exclusive_group()
    user_or_group.add_argument(
        '--user',
        metavar='<user>',
        help=_('Include <user> (name or ID)'),
    )
    user_or_group.add_argument(
        '--group',
        metavar='<group>',
        help=_('Include <group> (name or ID)'),
    )
    common.add_group_domain_option_to_parser(parser)
    common.add_project_domain_option_to_parser(parser)
    common.add_user_domain_option_to_parser(parser)
    common.add_inherited_option_to_parser(parser)


def _process_identity_and_resource_options(
    parsed_args, identity_client, validate_actor_existence=True
):
    def _find_user():
        domain_id = (
            common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.user_domain,
                validate_actor_existence=validate_actor_existence,
            )
            if parsed_args.user_domain
            else None
        )
        return common._find_sdk_id(
            identity_client.find_user,
            name_or_id=parsed_args.user,
            validate_actor_existence=validate_actor_existence,
            domain_id=domain_id,
        )

    def _find_group():
        domain_id = (
            common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.group_domain,
                validate_actor_existence=validate_actor_existence,
            )
            if parsed_args.group_domain
            else None
        )
        return common._find_sdk_id(
            identity_client.find_group,
            name_or_id=parsed_args.group,
            validate_actor_existence=validate_actor_existence,
            domain_id=domain_id,
        )

    def _find_project():
        domain_id = (
            common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.project_domain,
                validate_actor_existence=validate_actor_existence,
            )
            if parsed_args.project_domain
            else None
        )
        return common._find_sdk_id(
            identity_client.find_project,
            name_or_id=parsed_args.project,
            validate_actor_existence=validate_actor_existence,
            domain_id=domain_id,
        )

    kwargs = {}
    if parsed_args.user and parsed_args.system:
        kwargs['user'] = _find_user()
        kwargs['system'] = parsed_args.system
    elif parsed_args.user and parsed_args.domain:
        kwargs['user'] = _find_user()
        kwargs['domain'] = common._find_sdk_id(
            identity_client.find_domain,
            name_or_id=parsed_args.domain,
            validate_actor_existence=validate_actor_existence,
        )
    elif parsed_args.user and parsed_args.project:
        kwargs['user'] = _find_user()
        kwargs['project'] = _find_project()
    elif parsed_args.group and parsed_args.system:
        kwargs['group'] = _find_group()
        kwargs['system'] = parsed_args.system
    elif parsed_args.group and parsed_args.domain:
        kwargs['group'] = _find_group()
        kwargs['domain'] = common._find_sdk_id(
            identity_client.find_domain,
            name_or_id=parsed_args.domain,
            validate_actor_existence=validate_actor_existence,
        )
    elif parsed_args.group and parsed_args.project:
        kwargs['group'] = _find_group()
        kwargs['project'] = _find_project()
    else:
        msg = _(
            "Role not added, incorrect set of arguments "
            "provided. See openstack --help for more details"
        )
        raise exceptions.CommandError(msg)

    kwargs['inherited'] = parsed_args.inherited
    return kwargs


class AddRole(command.Command):
    _description = _(
        "Adds a role assignment to a user or group on the "
        "system, a domain, or a project"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to add to <user> (name or ID)'),
        )
        _add_identity_and_resource_options_to_parser(parser)
        common.add_role_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        if (
            not parsed_args.user
            and not parsed_args.domain
            and not parsed_args.group
            and not parsed_args.project
        ):
            msg = _(
                "Role not added, incorrect set of arguments "
                "provided. See openstack --help for more details"
            )
            raise exceptions.CommandError(msg)

        domain_id = None
        if parsed_args.role_domain:
            domain_id = common._find_sdk_id(
                identity_client.find_domain, name_or_id=parsed_args.role_domain
            )
        role = common._find_sdk_id(
            identity_client.find_role,
            name_or_id=parsed_args.role,
            domain_id=domain_id,
        )

        add_kwargs = _process_identity_and_resource_options(
            parsed_args, identity_client
        )

        if add_kwargs.get("domain"):
            if add_kwargs.get("user"):
                identity_client.assign_domain_role_to_user(
                    domain=add_kwargs["domain"],
                    user=add_kwargs["user"],
                    role=role,
                    inherited=add_kwargs["inherited"],
                )
            if add_kwargs.get("group"):
                identity_client.assign_domain_role_to_group(
                    domain=add_kwargs["domain"],
                    group=add_kwargs["group"],
                    role=role,
                    inherited=add_kwargs["inherited"],
                )
        elif add_kwargs.get("project"):
            if add_kwargs.get("user"):
                identity_client.assign_project_role_to_user(
                    project=add_kwargs["project"],
                    user=add_kwargs["user"],
                    role=role,
                    inherited=add_kwargs["inherited"],
                )
            if add_kwargs.get("group"):
                identity_client.assign_project_role_to_group(
                    project=add_kwargs["project"],
                    group=add_kwargs["group"],
                    role=role,
                    inherited=add_kwargs["inherited"],
                )
        elif add_kwargs.get("system"):
            if add_kwargs["inherited"]:
                LOG.warning(
                    _(
                        "'--inherited' was given, which is not supported "
                        "when adding a system role. This will be an error "
                        "in a future release."
                    )
                )
                # TODO(0weng): This should be an error in a future release
            if add_kwargs.get("user"):
                identity_client.assign_system_role_to_user(
                    system=add_kwargs["system"],
                    user=add_kwargs["user"],
                    role=role,
                )
            if add_kwargs.get("group"):
                identity_client.assign_system_role_to_group(
                    system=add_kwargs["system"],
                    group=add_kwargs["group"],
                    role=role,
                )


class CreateRole(command.ShowOne):
    _description = _("Create new role")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<role-name>',
            help=_('New role name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Add description about the role'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            help=_('Return existing role'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        create_kwargs = {}
        if parsed_args.domain:
            create_kwargs['domain_id'] = common._find_sdk_id(
                identity_client.find_domain, name_or_id=parsed_args.domain
            )

        if parsed_args.name:
            create_kwargs['name'] = parsed_args.name

        if parsed_args.description:
            create_kwargs['description'] = parsed_args.description

        if parsed_args.immutable is not None:
            create_kwargs['options'] = {"immutable": parsed_args.immutable}

        try:
            role = identity_client.create_role(**create_kwargs)

        except sdk_exc.ConflictException:
            if parsed_args.or_show:
                role = identity_client.find_role(
                    name_or_id=parsed_args.name,
                    domain_id=parsed_args.domain,
                    ignore_missing=False,
                )
                LOG.info(_('Returning existing role %s'), role.name)
            else:
                raise

        return _format_role(role)


class DeleteRole(command.Command):
    _description = _("Delete role(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'roles',
            metavar='<role>',
            nargs='+',
            help=_('Role(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common._find_sdk_id(
                identity_client.find_domain, parsed_args.domain
            )
        errors = 0
        for role in parsed_args.roles:
            try:
                role_id = common._find_sdk_id(
                    identity_client.find_role,
                    name_or_id=role,
                    domain_id=domain_id,
                )
                identity_client.delete_role(role=role_id, ignore_missing=False)
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete role with "
                        "name or ID '%(role)s': %(e)s"
                    ),
                    {'role': role, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.roles)
            msg = _("%(errors)s of %(total)s roles failed to delete.") % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListRole(command.Lister):
    _description = _("List roles")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Include <domain> (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        if parsed_args.domain:
            domain = identity_client.find_domain(
                name_or_id=parsed_args.domain,
                ignore_missing=False,
            )
            data = identity_client.roles(domain_id=domain.id)
            return (
                ('ID', 'Name', 'Domain'),
                (
                    utils.get_item_properties(s, ('id', 'name'))
                    + (domain.name,)
                    for s in data
                ),
            )

        else:
            data = identity_client.roles()
            return (
                ('ID', 'Name'),
                (utils.get_item_properties(s, ('id', 'name')) for s in data),
            )


class RemoveRole(command.Command):
    _description = _(
        "Removes a role assignment from system/domain/project : user/group"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to remove (name or ID)'),
        )
        _add_identity_and_resource_options_to_parser(parser)
        common.add_role_domain_option_to_parser(parser)

        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        if (
            not parsed_args.user
            and not parsed_args.domain
            and not parsed_args.group
            and not parsed_args.project
        ):
            msg = _(
                "Incorrect set of arguments provided. "
                "See openstack --help for more details"
            )
            raise exceptions.CommandError(msg)

        domain_id = None
        if parsed_args.role_domain:
            domain_id = common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.role_domain,
            )
        role = common._find_sdk_id(
            identity_client.find_role,
            name_or_id=parsed_args.role,
            domain_id=domain_id,
        )

        remove_kwargs = _process_identity_and_resource_options(
            parsed_args,
            identity_client,
            validate_actor_existence=False,
        )

        if remove_kwargs.get("domain"):
            if remove_kwargs.get("user"):
                identity_client.unassign_domain_role_from_user(
                    domain=remove_kwargs["domain"],
                    user=remove_kwargs["user"],
                    role=role,
                    inherited=remove_kwargs["inherited"],
                )
            if remove_kwargs.get("group"):
                identity_client.unassign_domain_role_from_group(
                    domain=remove_kwargs["domain"],
                    group=remove_kwargs["group"],
                    role=role,
                    inherited=remove_kwargs["inherited"],
                )
        elif remove_kwargs.get("project"):
            if remove_kwargs.get("user"):
                identity_client.unassign_project_role_from_user(
                    project=remove_kwargs["project"],
                    user=remove_kwargs["user"],
                    role=role,
                    inherited=remove_kwargs["inherited"],
                )
            if remove_kwargs.get("group"):
                identity_client.unassign_project_role_from_group(
                    project=remove_kwargs["project"],
                    group=remove_kwargs["group"],
                    role=role,
                    inherited=remove_kwargs["inherited"],
                )
        elif remove_kwargs.get("system"):
            if remove_kwargs.get("user"):
                identity_client.unassign_system_role_from_user(
                    system=remove_kwargs["system"],
                    user=remove_kwargs["user"],
                    role=role,
                )
            if remove_kwargs.get("group"):
                identity_client.unassign_system_role_from_group(
                    system=remove_kwargs["system"],
                    group=remove_kwargs["group"],
                    role=role,
                )


class SetRole(command.Command):
    _description = _("Set role properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to modify (name or ID)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Add description about the role'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set role name'),
        )
        common.add_resource_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        update_kwargs = {}
        if parsed_args.description:
            update_kwargs["description"] = parsed_args.description
        if parsed_args.name:
            update_kwargs["name"] = parsed_args.name

        domain_id = None
        if parsed_args.domain:
            domain_id = common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.domain,
            )
            update_kwargs["domain_id"] = domain_id

        if parsed_args.immutable is not None:
            update_kwargs["options"] = {"immutable": parsed_args.immutable}

        role = common._find_sdk_id(
            identity_client.find_role,
            name_or_id=parsed_args.role,
            domain_id=domain_id,
        )
        update_kwargs["role"] = role

        identity_client.update_role(**update_kwargs)


class ShowRole(command.ShowOne):
    _description = _("Display role details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'role',
            metavar='<role>',
            help=_('Role to display (name or ID)'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the role belongs to (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity

        domain_id = None
        if parsed_args.domain:
            domain_id = common._find_sdk_id(
                identity_client.find_domain,
                name_or_id=parsed_args.domain,
            )

        role = identity_client.find_role(
            name_or_id=parsed_args.role,
            domain_id=domain_id,
            ignore_missing=False,
        )

        return _format_role(role)



#   Copyright 2020 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import getpass
import logging
import os
import queue
import typing as ty

from cliff.formatters import table

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


def ask_user_yesno(msg):
    """Ask user Y/N question

    :param str msg: question text
    :return bool: User choice
    """
    while True:
        answer = getpass.getpass('{} [{}]: '.format(msg, 'y/n'))
        if answer in ('y', 'Y', 'yes'):
            return True
        elif answer in ('n', 'N', 'no'):
            return False


class ProjectCleanup(command.Command):
    _description = _("Clean resources associated with a project")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        action_group = parser.add_mutually_exclusive_group()
        action_group.add_argument(
            '--dry-run',
            action='store_true',
            help=_("List a project's resources but do not delete them"),
        )
        action_group.add_argument(
            '--auto-approve',
            action='store_true',
            help=_("Delete resources without asking for confirmation"),
        )
        project_group = parser.add_mutually_exclusive_group(required=True)
        project_group.add_argument(
            '--auth-project',
            action='store_true',
            help=_('Delete resources of the project used to authenticate'),
        )
        project_group.add_argument(
            '--project',
            metavar='<project>',
            help=_('Project to clean (name or ID)'),
        )
        parser.add_argument(
            '--created-before',
            metavar='<YYYY-MM-DDTHH24:MI:SS>',
            help=_('Only delete resources created before the given time'),
        )
        parser.add_argument(
            '--updated-before',
            metavar='<YYYY-MM-DDTHH24:MI:SS>',
            help=_('Only delete resources updated before the given time'),
        )
        parser.add_argument(
            '--skip-resource',
            metavar='<resource>',
            help='Skip cleanup of specific resource (repeat if necessary)',
            action='append',
        )
        identity_common.add_project_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        connection = self.app.client_manager.sdk_connection

        if parsed_args.auth_project:
            # is we've got a project already configured, use the connection
            # as-is
            pass
        elif parsed_args.project:
            project = connection.identity.find_project(
                name_or_id=parsed_args.project, ignore_missing=False
            )
            connection = connection.connect_as_project(project)

        if connection:
            status_queue: queue.Queue[ty.Any] = queue.Queue()
            parsed_args.max_width = int(
                os.environ.get('CLIFF_MAX_TERM_WIDTH', 0)
            )
            parsed_args.fit_width = bool(
                int(os.environ.get('CLIFF_FIT_WIDTH', 0))
            )
            parsed_args.print_empty = False
            table_fmt = table.TableFormatter()

            self.log.info('Searching resources...')

            filters = {}
            if parsed_args.created_before:
                filters['created_at'] = parsed_args.created_before

            if parsed_args.updated_before:
                filters['updated_at'] = parsed_args.updated_before

            connection.project_cleanup(
                dry_run=True,
                status_queue=status_queue,
                filters=filters,
                skip_resources=parsed_args.skip_resource,
            )

            data = []
            while not status_queue.empty():
                resource = status_queue.get_nowait()
                data.append(
                    (type(resource).__name__, resource.id, resource.name)
                )
                status_queue.task_done()
            status_queue.join()
            table_fmt.emit_list(
                ('Type', 'ID', 'Name'), data, self.app.stdout, parsed_args
            )

            if parsed_args.dry_run:
                return

            if not parsed_args.auto_approve:
                if not ask_user_yesno(
                    _("These resources will be deleted. Are you sure")
                ):
                    return

            self.log.warning(_('Deleting resources'))

            connection.project_cleanup(
                dry_run=False,
                status_queue=status_queue,
                filters=filters,
                skip_resources=parsed_args.skip_resource,
            )



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions

from openstackclient.image.v2 import metadef_properties
from openstackclient.tests.unit.image.v2 import fakes as image_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestMetadefPropertyCreate(image_fakes.TestImagev2):
    _metadef_namespace = image_fakes.create_one_metadef_namespace()
    _metadef_property = image_fakes.create_one_metadef_property()
    expected_columns = (
        'name',
        'title',
        'type',
    )
    expected_data = (
        _metadef_property.name,
        _metadef_property.title,
        _metadef_property.type,
    )

    def setUp(self):
        super().setUp()
        self.image_client.create_metadef_property.return_value = (
            self._metadef_property
        )
        self.cmd = metadef_properties.CreateMetadefProperty(self.app, None)

    def test_metadef_property_create_missing_arguments(self):
        self.assertRaises(
            tests_utils.ParserException, self.check_parser, self.cmd, [], []
        )

    def test_metadef_property_create(self):
        arglist = [
            '--name',
            'cpu_cores',
            '--schema',
            '{}',
            '--title',
            'vCPU Cores',
            '--type',
            'integer',
            self._metadef_namespace.namespace,
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_metadef_property_create_invalid_schema(self):
        arglist = [
            '--name',
            'cpu_cores',
            '--schema',
            '{invalid}',
            '--title',
            'vCPU Cores',
            '--type',
            'integer',
            self._metadef_namespace.namespace,
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestMetadefPropertyDelete(image_fakes.TestImagev2):
    def setUp(self):
        super().setUp()

        self.cmd = metadef_properties.DeleteMetadefProperty(self.app, None)
        self.image_client.delete_all_metadef_properties.return_value = None

    def test_metadef_property_delete(self):
        arglist = ['namespace', 'property']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.image_client.delete_metadef_property.assert_called_with(
            'property', 'namespace', ignore_missing=False
        )
        self.image_client.delete_all_metadef_properties.assert_not_called()

    def test_metadef_property_delete_missing_arguments(self):
        arglist = []
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            [],
        )
        self.image_client.delete_all_metadef_properties.assert_not_called()
        self.image_client.delete_metadef_property.assert_not_called()

    def test_metadef_property_delete_exception(self):
        arglist = ['namespace', 'property']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.image_client.delete_metadef_property.side_effect = (
            sdk_exceptions.ResourceNotFound
        )

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.image_client.delete_metadef_property.assert_called_with(
            'property', 'namespace', ignore_missing=False
        )
        self.image_client.delete_all_metadef_properties.assert_not_called()

    def test_metadef_property_delete_all(self):
        arglist = ['namespace']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.image_client.delete_all_metadef_properties.assert_called_with(
            'namespace'
        )
        self.image_client.delete_metadef_property.assert_not_called()


class TestMetadefPropertyList(image_fakes.TestImagev2):
    _metadef_property = [image_fakes.create_one_metadef_property()]

    columns = ['name', 'title', 'type']

    def setUp(self):
        super().setUp()

        self.image_client.metadef_properties.side_effect = [
            self._metadef_property,
            [],
        ]

        self.cmd = metadef_properties.ListMetadefProperties(self.app, None)
        self.datalist = self._metadef_property

    def test_metadef_property_list(self):
        arglist = ['my-namespace']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(getattr(self.datalist[0], 'name'), next(data)[0])


class TestMetadefPropertySet(image_fakes.TestImagev2):
    _metadef_property = image_fakes.create_one_metadef_property()

    def setUp(self):
        super().setUp()

        self.cmd = metadef_properties.SetMetadefProperty(self.app, None)
        self.image_client.get_metadef_property.return_value = (
            self._metadef_property
        )

    def test_metadef_property_set(self):
        arglist = ['--title', 'new title', 'namespace', 'property']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

    def test_metadef_property_set_invalid_schema(self):
        arglist = [
            '--title',
            'new title',
            '--schema',
            '{invalid}',
            'namespace',
            'property',
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestMetadefPropertyShow(image_fakes.TestImagev2):
    _metadef_property = image_fakes.create_one_metadef_property()

    expected_columns = (
        'name',
        'title',
        'type',
    )
    expected_data = (
        _metadef_property.name,
        _metadef_property.title,
        _metadef_property.type,
    )

    def setUp(self):
        super().setUp()

        self.image_client.get_metadef_property.return_value = (
            self._metadef_property
        )

        self.cmd = metadef_properties.ShowMetadefProperty(self.app, None)

    def test_metadef_property_show(self):
        arglist = ['my-namespace', 'my-property']
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Floating IP Pool action implementations"""

from osc_lib import exceptions

from openstackclient.api import compute_v2
from openstackclient.i18n import _
from openstackclient.network import common


class ListFloatingIPPool(common.NetworkAndComputeLister):
    _description = _("List pools of floating IP addresses")

    def take_action_network(self, client, parsed_args):
        msg = _(
            "Floating ip pool operations are only available for "
            "Compute v2 network."
        )
        raise exceptions.CommandError(msg)

    def take_action_compute(self, client, parsed_args):
        columns = ('Name',)
        data = [
            (x['name'],) for x in compute_v2.list_floating_ip_pools(client)
        ]

        return (columns, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cinderclient import api_versions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


def _format_cluster(cluster, detailed=False):
    columns: tuple[str, ...] = (
        'name',
        'binary',
        'state',
        'status',
    )
    column_headers: tuple[str, ...] = (
        'Name',
        'Binary',
        'State',
        'Status',
    )

    if detailed:
        columns += (
            'disabled_reason',
            'num_hosts',
            'num_down_hosts',
            'last_heartbeat',
            'created_at',
            'updated_at',
            # optional columns, depending on whether replication is enabled
            'replication_status',
            'frozen',
            'active_backend_id',
        )
        column_headers += (
            'Disabled Reason',
            'Hosts',
            'Down Hosts',
            'Last Heartbeat',
            'Created At',
            'Updated At',
            # optional columns, depending on whether replication is enabled
            'Replication Status',
            'Frozen',
            'Active Backend ID',
        )

    return (
        column_headers,
        utils.get_item_properties(
            cluster,
            columns,
        ),
    )


class ListBlockStorageCluster(command.Lister):
    """List block storage clusters.

    This command requires ``--os-volume-api-version`` 3.7 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--cluster',
            metavar='<name>',
            default=None,
            help=_(
                'Filter by cluster name, without backend will list '
                'all clustered services from the same cluster.'
            ),
        )
        parser.add_argument(
            '--binary',
            metavar='<binary>',
            help=_('Cluster binary.'),
        )
        parser.add_argument(
            '--up',
            action='store_true',
            dest='is_up',
            default=None,
            help=_('Filter by up status.'),
        )
        parser.add_argument(
            '--down',
            action='store_false',
            dest='is_up',
            help=_('Filter by down status.'),
        )
        parser.add_argument(
            '--disabled',
            action='store_true',
            dest='is_disabled',
            default=None,
            help=_('Filter by disabled status.'),
        )
        parser.add_argument(
            '--enabled',
            action='store_false',
            dest='is_disabled',
            help=_('Filter by enabled status.'),
        )
        parser.add_argument(
            '--num-hosts',
            metavar='<hosts>',
            type=int,
            default=None,
            help=_('Filter by number of hosts in the cluster.'),
        )
        parser.add_argument(
            '--num-down-hosts',
            metavar='<hosts>',
            type=int,
            default=None,
            help=_('Filter by number of hosts that are down.'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.7'):
            msg = _(
                "--os-volume-api-version 3.7 or greater is required to "
                "support the 'block storage cluster list' command"
            )
            raise exceptions.CommandError(msg)

        columns: tuple[str, ...] = ('Name', 'Binary', 'State', 'Status')
        if parsed_args.long:
            columns += (
                'Num Hosts',
                'Num Down Hosts',
                'Last Heartbeat',
                'Disabled Reason',
                'Created At',
                'Updated At',
            )

        data = volume_client.clusters.list(
            name=parsed_args.cluster,
            binary=parsed_args.binary,
            is_up=parsed_args.is_up,
            disabled=parsed_args.is_disabled,
            num_hosts=parsed_args.num_hosts,
            num_down_hosts=parsed_args.num_down_hosts,
            detailed=parsed_args.long,
        )

        return (
            columns,
            (utils.get_item_properties(s, columns) for s in data),
        )


class SetBlockStorageCluster(command.Command):
    """Set block storage cluster properties.

    This command requires ``--os-volume-api-version`` 3.7 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name of block storage cluster to update (name only)'),
        )
        parser.add_argument(
            '--binary',
            metavar='<binary>',
            default='cinder-volume',
            help=_(
                "Name of binary to filter by; defaults to 'cinder-volume' "
                "(optional)"
            ),
        )
        enabled_group = parser.add_mutually_exclusive_group()
        enabled_group.add_argument(
            '--enable',
            action='store_false',
            dest='disabled',
            default=None,
            help=_('Enable cluster'),
        )
        enabled_group.add_argument(
            '--disable',
            action='store_true',
            dest='disabled',
            help=_('Disable cluster'),
        )
        parser.add_argument(
            '--disable-reason',
            metavar='<reason>',
            dest='disabled_reason',
            help=_(
                'Reason for disabling the cluster '
                '(should be used with --disable option)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.7'):
            msg = _(
                "--os-volume-api-version 3.7 or greater is required to "
                "support the 'block storage cluster set' command"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.disabled_reason and not parsed_args.disabled:
            msg = _("Cannot specify --disable-reason without --disable")
            raise exceptions.CommandError(msg)

        cluster = volume_client.clusters.update(
            parsed_args.cluster,
            parsed_args.binary,
            disabled=parsed_args.disabled,
            disabled_reason=parsed_args.disabled_reason,
        )

        return _format_cluster(cluster, detailed=True)


class ShowBlockStorageCluster(command.ShowOne):
    """Show detailed information for a block storage cluster.

    This command requires ``--os-volume-api-version`` 3.7 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name of block storage cluster.'),
        )
        parser.add_argument(
            '--binary',
            metavar='<binary>',
            help=_('Service binary.'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if volume_client.api_version < api_versions.APIVersion('3.7'):
            msg = _(
                "--os-volume-api-version 3.7 or greater is required to "
                "support the 'block storage cluster show' command"
            )
            raise exceptions.CommandError(msg)

        cluster = volume_client.clusters.show(
            parsed_args.cluster,
            binary=parsed_args.binary,
        )

        return _format_cluster(cluster, detailed=True)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib import exceptions

from openstackclient.i18n import _


# Transform compute security group rule for display.
def transform_compute_security_group_rule(sg_rule):
    info = {}
    info.update(sg_rule)
    from_port = info.pop('from_port')
    to_port = info.pop('to_port')
    if isinstance(from_port, int) and isinstance(to_port, int):
        port_range = {'port_range': f"{from_port}:{to_port}"}
    elif from_port is None and to_port is None:
        port_range = {'port_range': ""}
    else:
        port_range = {'port_range': f"{from_port}:{to_port}"}
    info.update(port_range)
    if 'cidr' in info['ip_range']:
        info['ip_range'] = info['ip_range']['cidr']
    else:
        info['ip_range'] = ''
    if info['ip_protocol'] is None:
        info['ip_protocol'] = ''
    elif info['ip_protocol'].lower() == 'icmp':
        info['port_range'] = ''
    group = info.pop('group')
    if 'name' in group:
        info['remote_security_group'] = group['name']
    else:
        info['remote_security_group'] = ''
    return info


def str2bool(strbool):
    if strbool is None:
        return None
    return strbool.lower() == 'true'


def str2list(strlist):
    result = []
    if strlist:
        result = strlist.split(';')
    return result


def str2dict(strdict: str) -> dict[str, str]:
    """Convert key1:value1;key2:value2;... string into dictionary.

    :param strdict: string in the form of key1:value1;key2:value2
    """
    result: dict[str, str] = {}
    if not strdict:
        return result
    i = 0
    kvlist = []
    for kv in strdict.split(';'):
        if ':' in kv:
            kvlist.append(kv)
            i += 1
        elif i == 0:
            msg = _("missing value for key '%s'")
            raise exceptions.CommandError(msg % kv)
        else:
            kvlist[i - 1] = f"{kvlist[i - 1]};{kv}"
    for kv in kvlist:
        key, sep, value = kv.partition(':')
        result[key] = value
    return result


def format_security_group_rule_show(obj):
    data = transform_compute_security_group_rule(obj)
    return zip(*sorted(data.items()))


def format_network_port_range(rule):
    # Display port range or ICMP type and code. For example:
    # - ICMP type: 'type=3'
    # - ICMP type and code: 'type=3:code=0'
    # - ICMP code: Not supported
    # - Matching port range: '443:443'
    # - Different port range: '22:24'
    # - Single port: '80:80'
    # - No port range: ''
    port_range = ''
    if is_icmp_protocol(rule['protocol']):
        if rule['port_range_min']:
            port_range += 'type=' + str(rule['port_range_min'])
        if rule['port_range_max']:
            port_range += ':code=' + str(rule['port_range_max'])
    elif rule['port_range_min'] or rule['port_range_max']:
        port_range_min = str(rule['port_range_min'])
        port_range_max = str(rule['port_range_max'])
        if rule['port_range_min'] is None:
            port_range_min = port_range_max
        if rule['port_range_max'] is None:
            port_range_max = port_range_min
        port_range = port_range_min + ':' + port_range_max
    return port_range


def format_remote_ip_prefix(rule):
    remote_ip_prefix = rule['remote_ip_prefix']
    if remote_ip_prefix is None:
        ethertype = rule['ether_type']
        if ethertype == 'IPv4':
            remote_ip_prefix = '0.0.0.0/0'
        elif ethertype == 'IPv6':
            remote_ip_prefix = '::/0'
    return remote_ip_prefix


def convert_ipvx_case(string):
    if string.lower() == 'ipv4':
        return 'IPv4'
    if string.lower() == 'ipv6':
        return 'IPv6'
    return string


def is_icmp_protocol(protocol):
    # NOTE(rtheis): Neutron has deprecated protocol icmpv6.
    # However, while the OSC CLI doesn't document the protocol,
    # the code must still handle it. In addition, handle both
    # protocol names and numbers.
    if protocol in ['icmp', 'icmpv6', 'ipv6-icmp', '1', '58']:
        return True
    else:
        return False


def convert_to_lowercase(string):
    return string.lower()


def get_protocol(parsed_args, default_protocol='any'):
    protocol = default_protocol
    if parsed_args.protocol is not None:
        protocol = parsed_args.protocol
    if hasattr(parsed_args, "proto") and parsed_args.proto is not None:
        protocol = parsed_args.proto
    if protocol == 'any':
        protocol = None
    return protocol


def get_ethertype(parsed_args, protocol):
    ethertype = 'IPv4'
    if parsed_args.ethertype is not None:
        ethertype = parsed_args.ethertype
    elif is_ipv6_protocol(protocol):
        ethertype = 'IPv6'
    return ethertype


def is_ipv6_protocol(protocol):
    # NOTE(rtheis): Neutron has deprecated protocol icmpv6.
    # However, while the OSC CLI doesn't document the protocol,
    # the code must still handle it. In addition, handle both
    # protocol names and numbers.
    if (
        protocol is not None
        and protocol.startswith('ipv6-')
        or protocol in ['icmpv6', '41', '43', '44', '58', '59', '60']
    ):
        return True
    else:
        return False



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Security Group action implementations"""

import argparse
import typing as ty

from cliff import columns as cliff_columns
from osc_lib import utils
from osc_lib.utils import tags as _tag

from openstackclient.api import compute_v2
from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common
from openstackclient.network import utils as network_utils


def _format_network_security_group_rules(sg_rules):
    # For readability and to align with formatting compute security group
    # rules, trim keys with caller known (e.g. security group and tenant ID)
    # or empty values.
    for sg_rule in sg_rules:
        empty_keys = [k for k, v in sg_rule.items() if not v]
        for key in empty_keys:
            sg_rule.pop(key)
        sg_rule.pop('security_group_id', None)
        sg_rule.pop('tenant_id', None)
        sg_rule.pop('project_id', None)
    return utils.format_list_of_dicts(sg_rules)


def _format_compute_security_group_rule(sg_rule):
    info = network_utils.transform_compute_security_group_rule(sg_rule)
    # Trim parent security group ID since caller has this information.
    info.pop('parent_group_id', None)
    # Trim keys with empty string values.
    keys_to_trim = [
        'ip_protocol',
        'ip_range',
        'port_range',
        'remote_security_group',
    ]
    for key in keys_to_trim:
        if key in info and not info[key]:
            info.pop(key)
    return utils.format_dict(info)


def _format_compute_security_group_rules(sg_rules):
    rules = []
    for sg_rule in sg_rules:
        rules.append(_format_compute_security_group_rule(sg_rule))
    return utils.format_list(rules, separator='\n')


class NetworkSecurityGroupRulesColumn(cliff_columns.FormattableColumn[ty.Any]):
    def human_readable(self):
        return _format_network_security_group_rules(self._value)


class ComputeSecurityGroupRulesColumn(cliff_columns.FormattableColumn[ty.Any]):
    def human_readable(self):
        return _format_compute_security_group_rules(self._value)


_formatters_network = {
    'security_group_rules': NetworkSecurityGroupRulesColumn,
}


_formatters_compute = {
    'rules': ComputeSecurityGroupRulesColumn,
}


def _get_columns(item):
    # We still support Nova managed security groups, where we have tenant_id.
    column_map = {
        'security_group_rules': 'rules',
    }
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


# TODO(abhiraut): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class CreateSecurityGroup(
    common.NetworkAndComputeShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create a new security group")

    def update_parser_common(self, parser):
        parser.add_argument(
            "name", metavar="<name>", help=_("New security group name")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Security group description"),
        )
        return parser

    def update_parser_network(self, parser):
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=self.enhance_help_neutron(_("Owner's project (name or ID)")),
        )
        stateful_group = parser.add_mutually_exclusive_group()
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            help=_("Security group is stateful (default)"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_true',
            default=None,
            help=_("Security group is stateless"),
        )
        identity_common.add_project_domain_option_to_parser(
            parser, enhance_help=self.enhance_help_neutron
        )
        _tag.add_tag_option_to_parser_for_create(
            parser, _('security group'), enhance_help=self.enhance_help_neutron
        )
        return parser

    def _get_description(self, parsed_args):
        if parsed_args.description is not None:
            return parsed_args.description
        else:
            return parsed_args.name

    def take_action_network(self, client, parsed_args):
        # Build the create attributes.
        attrs = {}
        attrs['name'] = parsed_args.name
        attrs['description'] = self._get_description(parsed_args)
        if parsed_args.stateful:
            attrs['stateful'] = True
        if parsed_args.stateless:
            attrs['stateful'] = False
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.identity
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            attrs['project_id'] = project_id
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        # Create the security group and display the results.
        obj = client.create_security_group(**attrs)
        # tags cannot be set when created, so tags need to be set later.
        _tag.update_tags_for_set(client, obj, parsed_args)
        display_columns, property_columns = _get_columns(obj)
        data = utils.get_item_properties(
            obj, property_columns, formatters=_formatters_network
        )
        return (display_columns, data)

    def take_action_compute(self, client, parsed_args):
        description = self._get_description(parsed_args)
        obj = compute_v2.create_security_group(
            client,
            parsed_args.name,
            description,
        )
        display_columns = ('description', 'id', 'name', 'project_id', 'rules')
        property_columns = ('description', 'id', 'name', 'tenant_id', 'rules')
        data = utils.get_dict_properties(
            obj, property_columns, formatters=_formatters_compute
        )
        return (display_columns, data)


class DeleteSecurityGroup(common.NetworkAndComputeDelete):
    _description = _("Delete security group(s)")

    # Used by base class to find resources in parsed_args.
    resource = 'group'
    r = None

    def update_parser_common(self, parser):
        parser.add_argument(
            'group',
            metavar='<group>',
            nargs="+",
            help=_("Security group(s) to delete (name or ID)"),
        )
        return parser

    def take_action_network(self, client, parsed_args):
        obj = client.find_security_group(self.r, ignore_missing=False)
        client.delete_security_group(obj)

    def take_action_compute(self, client, parsed_args):
        security_group = compute_v2.find_security_group(client, self.r)
        compute_v2.delete_security_group(client, security_group['id'])


# TODO(rauta): Use the SDK resource mapped attribute names once
# the OSC minimum requirements include SDK 1.0.
class ListSecurityGroup(common.NetworkAndComputeLister):
    _description = _("List security groups")
    FIELDS_TO_RETRIEVE = [
        'id',
        'name',
        'description',
        'project_id',
        'tags',
        'shared',
    ]

    def update_parser_network(self, parser):
        if not self.is_docs_build:
            # Maintain and hide the argument for backwards compatibility.
            # Network will always return all projects for an admin.
            parser.add_argument(
                '--all-projects',
                action='store_true',
                default=False,
                help=argparse.SUPPRESS,
            )

        parser.add_argument(
            '--project',
            metavar='<project>',
            help=self.enhance_help_neutron(
                _(
                    "List only security groups with the specified project "
                    "(name or ID)"
                )
            ),
        )
        identity_common.add_project_domain_option_to_parser(
            parser, enhance_help=self.enhance_help_neutron
        )

        shared_group = parser.add_mutually_exclusive_group()
        shared_group.add_argument(
            '--share',
            action='store_true',
            dest='shared',
            default=None,
            help=_("List only security groups shared between projects"),
        )
        shared_group.add_argument(
            '--no-share',
            action='store_false',
            dest='shared',
            default=None,
            help=_("List only security groups not shared between projects"),
        )

        _tag.add_tag_filtering_option_to_parser(
            parser, _('security group'), enhance_help=self.enhance_help_neutron
        )
        return parser

    def update_parser_compute(self, parser):
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=self.enhance_help_nova_network(
                _("Display information from all projects (admin only)")
            ),
        )
        return parser

    def take_action_network(self, client, parsed_args):
        filters = {}
        if parsed_args.project:
            identity_client = self.app.client_manager.identity
            project_id = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            filters['project_id'] = project_id

        if parsed_args.shared is not None:
            filters['shared'] = parsed_args.shared

        _tag.get_tag_filtering_args(parsed_args, filters)
        data = client.security_groups(
            fields=self.FIELDS_TO_RETRIEVE, **filters
        )

        columns = (
            "id",
            "name",
            "description",
            "project_id",
            "tags",
            "is_shared",
        )
        column_headers = (
            "ID",
            "Name",
            "Description",
            "Project",
            "Tags",
            "Shared",
        )
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )

    def take_action_compute(self, client, parsed_args):
        data = compute_v2.list_security_groups(
            # TODO(dtroyer): add limit, marker
            client,
            all_projects=parsed_args.all_projects,
        )

        columns: tuple[str, ...] = ("id", "name", "description")
        column_headers: tuple[str, ...] = ("ID", "Name", "Description")
        if parsed_args.all_projects:
            columns += ('tenant_id',)
            column_headers += ('Project',)
        return (
            column_headers,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class SetSecurityGroup(
    common.NetworkAndComputeCommand, common.NeutronCommandWithExtraArgs
):
    _description = _("Set security group properties")

    def update_parser_common(self, parser):
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_("Security group to modify (name or ID)"),
        )
        parser.add_argument(
            '--name', metavar='<new-name>', help=_("New security group name")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("New security group description"),
        )
        stateful_group = parser.add_mutually_exclusive_group()
        stateful_group.add_argument(
            "--stateful",
            action='store_true',
            default=None,
            help=_("Security group is stateful (default)"),
        )
        stateful_group.add_argument(
            "--stateless",
            action='store_true',
            default=None,
            help=_("Security group is stateless"),
        )
        return parser

    def update_parser_network(self, parser):
        _tag.add_tag_option_to_parser_for_set(
            parser, _('security group'), enhance_help=self.enhance_help_neutron
        )
        return parser

    def take_action_network(self, client, parsed_args):
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )
        attrs = {}
        if parsed_args.name is not None:
            attrs['name'] = parsed_args.name
        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description
        if parsed_args.stateful:
            attrs['stateful'] = True
        if parsed_args.stateless:
            attrs['stateful'] = False
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        # NOTE(rtheis): Previous behavior did not raise a CommandError
        # if there were no updates. Maintain this behavior and issue
        # the update.
        client.update_security_group(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_set(client, obj, parsed_args)

    def take_action_compute(self, client, parsed_args):
        security_group = compute_v2.find_security_group(
            client, parsed_args.group
        )

        params = {}
        if parsed_args.name is not None:
            params['name'] = parsed_args.name
        if parsed_args.description is not None:
            params['description'] = parsed_args.description

        # NOTE(rtheis): Previous behavior did not raise a CommandError
        # if there were no updates. Maintain this behavior and issue
        # the update.
        compute_v2.update_security_group(
            client, security_group['id'], **params
        )


class ShowSecurityGroup(common.NetworkAndComputeShowOne):
    _description = _("Display security group details")

    def update_parser_common(self, parser):
        parser.add_argument(
            'group',
            metavar='<group>',
            help=_("Security group to display (name or ID)"),
        )
        return parser

    def take_action_network(self, client, parsed_args):
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )
        display_columns, property_columns = _get_columns(obj)
        data = utils.get_item_properties(
            obj, property_columns, formatters=_formatters_network
        )
        return (display_columns, data)

    def take_action_compute(self, client, parsed_args):
        obj = compute_v2.find_security_group(client, parsed_args.group)
        display_columns = ('description', 'id', 'name', 'project_id', 'rules')
        property_columns = ('description', 'id', 'name', 'tenant_id', 'rules')
        data = utils.get_dict_properties(
            obj, property_columns, formatters=_formatters_compute
        )
        return (display_columns, data)


class UnsetSecurityGroup(command.Command):
    _description = _("Unset security group properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'group',
            metavar="<group>",
            help=_("Security group to modify (name or ID)"),
        )
        _tag.add_tag_option_to_parser_for_unset(parser, _('security group'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        obj = client.find_security_group(
            parsed_args.group, ignore_missing=False
        )

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_unset(client, obj, parsed_args)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class AddressScopeTests(common.NetworkTests):
    """Functional tests for address scope"""

    # NOTE(dtroyer): Do not normalize the setup and teardown of the resource
    #                creation and deletion.  Little is gained when each test
    #                has its own needs and there are collisions when running
    #                tests in parallel.

    def test_address_scope_delete(self):
        """Test create, delete multiple"""
        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'address scope create ' + name1,
            parse_output=True,
        )
        self.assertEqual(
            name1,
            cmd_output['name'],
        )
        # Check the default values
        self.assertFalse(cmd_output['shared'])

        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'address scope create ' + name2,
            parse_output=True,
        )
        self.assertEqual(
            name2,
            cmd_output['name'],
        )

        raw_output = self.openstack(
            'address scope delete ' + name1 + ' ' + name2,
        )
        self.assertOutput('', raw_output)

    def test_address_scope_list(self):
        """Test create defaults, list filters, delete"""
        name1 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'address scope create ' + '--ip-version 4 ' + '--share ' + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'address scope delete ' + name1)
        self.assertEqual(
            name1,
            cmd_output['name'],
        )
        self.assertEqual(
            4,
            cmd_output['ip_version'],
        )
        self.assertTrue(cmd_output['shared'])

        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'address scope create '
            + '--ip-version 6 '
            + '--no-share '
            + name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'address scope delete ' + name2)
        self.assertEqual(
            name2,
            cmd_output['name'],
        )
        self.assertEqual(
            6,
            cmd_output['ip_version'],
        )
        self.assertFalse(cmd_output['shared'])

        # Test list
        cmd_output = self.openstack(
            'address scope list ',
            parse_output=True,
        )
        col_data = [x["IP Version"] for x in cmd_output]
        self.assertIn(4, col_data)
        self.assertIn(6, col_data)

        # Test list --share
        cmd_output = self.openstack(
            'address scope list --share',
            parse_output=True,
        )
        col_data = [x["Shared"] for x in cmd_output]
        self.assertIn(True, col_data)
        self.assertNotIn(False, col_data)

        # Test list --no-share
        cmd_output = self.openstack(
            'address scope list --no-share',
            parse_output=True,
        )
        col_data = [x["Shared"] for x in cmd_output]
        self.assertIn(False, col_data)
        self.assertNotIn(True, col_data)

    def test_address_scope_set(self):
        """Tests create options, set, show, delete"""
        name = uuid.uuid4().hex
        newname = name + "_"
        cmd_output = self.openstack(
            'address scope create ' + '--ip-version 4 ' + '--no-share ' + name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'address scope delete ' + newname)
        self.assertEqual(
            name,
            cmd_output['name'],
        )
        self.assertEqual(
            4,
            cmd_output['ip_version'],
        )
        self.assertFalse(cmd_output['shared'])

        raw_output = self.openstack(
            'address scope set ' + '--name ' + newname + ' --share ' + name,
        )
        self.assertOutput('', raw_output)

        cmd_output = self.openstack(
            'address scope show ' + newname,
            parse_output=True,
        )
        self.assertEqual(
            newname,
            cmd_output['name'],
        )
        self.assertEqual(
            4,
            cmd_output['ip_version'],
        )
        self.assertTrue(cmd_output['shared'])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy

from requests_mock.contrib import fixture

from openstackclient.object.v1 import container as container_cmds
from openstackclient.tests.unit.object.v1 import fakes as object_fakes


class TestContainerAll(object_fakes.TestObjectv1):
    def setUp(self):
        super().setUp()

        self.requests_mock = self.useFixture(fixture.Fixture())


class TestContainerCreate(TestContainerAll):
    columns = (
        'account',
        'container',
        'x-trans-id',
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = container_cmds.CreateContainer(self.app, None)

    def test_object_create_container_single(self):
        self.requests_mock.register_uri(
            'PUT',
            object_fakes.ENDPOINT + '/ernie',
            headers={'x-trans-id': '314159'},
            status_code=200,
        )

        arglist = [
            'ernie',
        ]
        verifylist = [
            (
                'containers',
                ['ernie'],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (
                object_fakes.ACCOUNT_ID,
                'ernie',
                '314159',
            )
        ]
        self.assertEqual(datalist, list(data))

    def test_object_create_container_storage_policy(self):
        self.requests_mock.register_uri(
            'PUT',
            object_fakes.ENDPOINT + '/ernie',
            headers={'x-trans-id': '314159', 'x-storage-policy': 'o1--sr-r3'},
            status_code=200,
        )

        arglist = ['ernie', '--storage-policy', 'o1--sr-r3']
        verifylist = [
            ('containers', ['ernie']),
            ('storage_policy', 'o1--sr-r3'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (
                object_fakes.ACCOUNT_ID,
                'ernie',
                '314159',
            )
        ]
        self.assertEqual(datalist, list(data))

    def test_object_create_container_public(self):
        self.requests_mock.register_uri(
            'PUT',
            object_fakes.ENDPOINT + '/ernie',
            headers={
                'x-trans-id': '314159',
                'x-container-read': '.r:*,.rlistings',
            },
            status_code=200,
        )

        arglist = ['ernie', '--public']
        verifylist = [('containers', ['ernie']), ('public', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (
                object_fakes.ACCOUNT_ID,
                'ernie',
                '314159',
            )
        ]
        self.assertEqual(datalist, list(data))

    def test_object_create_container_more(self):
        self.requests_mock.register_uri(
            'PUT',
            object_fakes.ENDPOINT + '/ernie',
            headers={'x-trans-id': '314159'},
            status_code=200,
        )
        self.requests_mock.register_uri(
            'PUT',
            object_fakes.ENDPOINT + '/bert',
            headers={'x-trans-id': '42'},
            status_code=200,
        )

        arglist = [
            'ernie',
            'bert',
        ]
        verifylist = [
            (
                'containers',
                ['ernie', 'bert'],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (
                object_fakes.ACCOUNT_ID,
                'ernie',
                '314159',
            ),
            (
                object_fakes.ACCOUNT_ID,
                'bert',
                '42',
            ),
        ]
        self.assertEqual(datalist, list(data))


class TestContainerDelete(TestContainerAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = container_cmds.DeleteContainer(self.app, None)

    def test_object_delete_container_single(self):
        self.requests_mock.register_uri(
            'DELETE',
            object_fakes.ENDPOINT + '/ernie',
            status_code=200,
        )

        arglist = [
            'ernie',
        ]
        verifylist = [
            (
                'containers',
                ['ernie'],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Command.take_action() returns None
        ret = self.cmd.take_action(parsed_args)
        self.assertIsNone(ret)

    def test_object_delete_container_more(self):
        self.requests_mock.register_uri(
            'DELETE',
            object_fakes.ENDPOINT + '/ernie',
            status_code=200,
        )
        self.requests_mock.register_uri(
            'DELETE',
            object_fakes.ENDPOINT + '/bert',
            status_code=200,
        )

        arglist = [
            'ernie',
            'bert',
        ]
        verifylist = [
            (
                'containers',
                ['ernie', 'bert'],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Command.take_action() returns None
        ret = self.cmd.take_action(parsed_args)
        self.assertIsNone(ret)


class TestContainerList(TestContainerAll):
    columns = ('Name',)

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = container_cmds.ListContainer(self.app, None)

    def test_object_list_containers_no_options(self):
        return_body = [
            copy.deepcopy(object_fakes.CONTAINER),
            copy.deepcopy(object_fakes.CONTAINER_3),
            copy.deepcopy(object_fakes.CONTAINER_2),
        ]
        self.requests_mock.register_uri(
            'GET',
            object_fakes.ENDPOINT + '?format=json',
            json=return_body,
            status_code=200,
        )

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Lister.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (object_fakes.container_name,),
            (object_fakes.container_name_3,),
            (object_fakes.container_name_2,),
        ]
        self.assertEqual(datalist, list(data))

    def test_object_list_containers_prefix(self):
        return_body = [
            copy.deepcopy(object_fakes.CONTAINER),
            copy.deepcopy(object_fakes.CONTAINER_3),
        ]
        self.requests_mock.register_uri(
            'GET',
            object_fakes.ENDPOINT + '?format=json&prefix=bit',
            json=return_body,
            status_code=200,
        )

        arglist = [
            '--prefix',
            'bit',
        ]
        verifylist = [
            ('prefix', 'bit'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Lister.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (object_fakes.container_name,),
            (object_fakes.container_name_3,),
        ]
        self.assertEqual(datalist, list(data))


class TestContainerSave(TestContainerAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = container_cmds.SaveContainer(self.app, None)


# TODO(dtroyer): need to mock out object_lib.save_object() to test this
#     def test_object_save_container(self):
#         return_body = [
#             copy.deepcopy(object_fakes.OBJECT),
#             copy.deepcopy(object_fakes.OBJECT_2),
#         ]
#         # Initial container list request
#         self.requests_mock.register_uri(
#             'GET',
#             object_fakes.ENDPOINT + '/oscar?format=json',
#             json=return_body,
#             status_code=200,
#         )
#         # Individual object save requests
#         self.requests_mock.register_uri(
#             'GET',
#             object_fakes.ENDPOINT + '/oscar/' + object_fakes.object_name_1,
#             json=object_fakes.OBJECT,
#             status_code=200,
#         )
#         self.requests_mock.register_uri(
#             'GET',
#             object_fakes.ENDPOINT + '/oscar/' + object_fakes.object_name_2,
#             json=object_fakes.OBJECT_2,
#             status_code=200,
#         )
#
#         arglist = [
#             'oscar',
#         ]
#         verifylist = [(
#             'container', 'oscar',
#         )]
#         parsed_args = self.check_parser(self.cmd, arglist, verifylist)
#
#         # Command.take_action() returns None
#         ret = self.cmd.take_action(parsed_args)
#         self.assertIsNone(ret)


class TestContainerShow(TestContainerAll):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = container_cmds.ShowContainer(self.app, None)

    def test_object_show_container(self):
        headers = {
            'x-container-object-count': '42',
            'x-container-bytes-used': '123',
            'x-container-read': 'qaz',
            'x-container-write': 'wsx',
            'x-container-sync-to': 'edc',
            'x-container-sync-key': 'rfv',
            'x-storage-policy': 'o1--sr-r3',
        }
        self.requests_mock.register_uri(
            'HEAD',
            object_fakes.ENDPOINT + '/ernie',
            headers=headers,
            status_code=200,
        )

        arglist = [
            'ernie',
        ]
        verifylist = [
            (
                'container',
                'ernie',
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = (
            'account',
            'bytes_used',
            'container',
            'object_count',
            'read_acl',
            'storage_policy',
            'sync_key',
            'sync_to',
            'write_acl',
        )
        self.assertEqual(collist, columns)
        datalist = [
            object_fakes.ACCOUNT_ID,
            '123',
            'ernie',
            '42',
            'qaz',
            'o1--sr-r3',
            'rfv',
            'edc',
            'wsx',
        ]
        self.assertEqual(datalist, list(data))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 IdentityProvider action implementations"""

import logging

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


class CreateIdentityProvider(command.ShowOne):
    _description = _("Create new identity provider")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'identity_provider_id',
            metavar='<name>',
            help=_('New identity provider name (must be unique)'),
        )
        identity_remote_id_provider = parser.add_mutually_exclusive_group()
        identity_remote_id_provider.add_argument(
            '--remote-id',
            metavar='<remote-id>',
            dest='remote_ids',
            action='append',
            help=_(
                'Remote IDs to associate with the Identity Provider '
                '(repeat option to provide multiple values)'
            ),
        )
        identity_remote_id_provider.add_argument(
            '--remote-id-file',
            metavar='<file-name>',
            help=_(
                'Name of a file that contains many remote IDs to associate '
                'with the identity provider, one per line'
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New identity provider description'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_(
                'Domain to associate with the identity provider. If not '
                'specified, a domain will be created automatically. '
                '(Name or ID)'
            ),
        )
        parser.add_argument(
            '--authorization-ttl',
            metavar='<authorization-ttl>',
            type=int,
            help=_(
                'Time to keep the role assignments for users '
                'authenticating via this identity provider. '
                'When not provided, global default configured in the '
                'Identity service will be used. '
                'Available since Identity API version 3.14 (Ussuri).'
            ),
        )
        enable_identity_provider = parser.add_mutually_exclusive_group()
        enable_identity_provider.add_argument(
            '--enable',
            dest='enabled',
            action='store_true',
            default=True,
            help=_('Enable identity provider (default)'),
        )
        enable_identity_provider.add_argument(
            '--disable',
            dest='enabled',
            action='store_false',
            help=_('Disable the identity provider'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        remote_ids: list[str] | None = None
        if parsed_args.remote_id_file:
            file_content = utils.read_blob_file_contents(
                parsed_args.remote_id_file
            )
            remote_ids = file_content.splitlines()
            remote_ids = list(map(str.strip, remote_ids))
        elif parsed_args.remote_ids:
            remote_ids = parsed_args.remote_ids

        domain_id = None
        if parsed_args.domain:
            domain_id = common.find_domain(
                identity_client, parsed_args.domain
            ).id

        # TODO(pas-ha) actually check for 3.14 microversion
        kwargs = {}
        auth_ttl = parsed_args.authorization_ttl
        if auth_ttl is not None:
            if auth_ttl < 0:
                msg = _("%(param)s must be positive integer or zero.") % {
                    "param": "authorization-ttl"
                }
                raise exceptions.CommandError(msg)
            kwargs['authorization_ttl'] = auth_ttl

        idp = identity_client.federation.identity_providers.create(
            id=parsed_args.identity_provider_id,
            remote_ids=remote_ids,
            description=parsed_args.description,
            domain_id=domain_id,
            enabled=parsed_args.enabled,
            **kwargs,
        )

        idp._info.pop('links', None)
        idp._info['remote_ids'] = format_columns.ListColumn(
            idp._info.pop('remote_ids', [])
        )
        return zip(*sorted(idp._info.items()))


class DeleteIdentityProvider(command.Command):
    _description = _("Delete identity provider(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'identity_provider',
            metavar='<identity-provider>',
            nargs='+',
            help=_('Identity provider(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        result = 0
        for i in parsed_args.identity_provider:
            try:
                identity_client.federation.identity_providers.delete(i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete identity providers with "
                        "name or ID '%(provider)s': %(e)s"
                    ),
                    {'provider': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.identity_provider)
            msg = _(
                "%(result)s of %(total)s identity providers failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListIdentityProvider(command.Lister):
    _description = _("List identity providers")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--id',
            metavar='<id>',
            help=_('The Identity Providers’ ID attribute'),
        )
        parser.add_argument(
            '--enabled',
            dest='enabled',
            action='store_true',
            help=_('The Identity Providers that are enabled will be returned'),
        )
        return parser

    def take_action(self, parsed_args):
        columns = ('ID', 'Enabled', 'Domain ID', 'Description')
        identity_client = self.app.client_manager.identity

        kwargs = {}
        if parsed_args.id:
            kwargs['id'] = parsed_args.id
        if parsed_args.enabled:
            kwargs['enabled'] = True

        data = identity_client.federation.identity_providers.list(**kwargs)
        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetIdentityProvider(command.Command):
    _description = _("Set identity provider properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'identity_provider',
            metavar='<identity-provider>',
            help=_('Identity provider to modify'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set identity provider description'),
        )
        identity_remote_id_provider = parser.add_mutually_exclusive_group()
        identity_remote_id_provider.add_argument(
            '--remote-id',
            metavar='<remote-id>',
            dest='remote_ids',
            action='append',
            help=_(
                'Remote IDs to associate with the Identity Provider '
                '(repeat option to provide multiple values)'
            ),
        )
        identity_remote_id_provider.add_argument(
            '--remote-id-file',
            metavar='<file-name>',
            help=_(
                'Name of a file that contains many remote IDs to associate '
                'with the identity provider, one per line'
            ),
        )
        parser.add_argument(
            '--authorization-ttl',
            metavar='<authorization-ttl>',
            type=int,
            help=_(
                'Time to keep the role assignments for users '
                'authenticating via this identity provider. '
                'Available since Identity API version 3.14 (Ussuri).'
            ),
        )
        enable_identity_provider = parser.add_mutually_exclusive_group()
        enable_identity_provider.add_argument(
            '--enable',
            action='store_true',
            help=_('Enable the identity provider'),
        )
        enable_identity_provider.add_argument(
            '--disable',
            action='store_true',
            help=_('Disable the identity provider'),
        )
        return parser

    def take_action(self, parsed_args):
        federation_client = self.app.client_manager.identity.federation

        # Always set remote_ids if either is passed in
        if parsed_args.remote_id_file:
            file_content = utils.read_blob_file_contents(
                parsed_args.remote_id_file
            )
            remote_ids = file_content.splitlines()
            remote_ids = list(map(str.strip, remote_ids))
        elif parsed_args.remote_ids:
            remote_ids = parsed_args.remote_ids

        # Setup keyword args for the client
        kwargs = {}
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if parsed_args.enable:
            kwargs['enabled'] = True
        if parsed_args.disable:
            kwargs['enabled'] = False
        if parsed_args.remote_id_file or parsed_args.remote_ids:
            kwargs['remote_ids'] = remote_ids

        # TODO(pas-ha) actually check for 3.14 microversion
        # TODO(pas-ha) make it possible to reset authorization_ttl
        # back to None value.
        # Currently not possible as filter_kwargs decorator in
        # keystoneclient/base.py explicitly drops the None-valued keys
        # from kwargs, and 'update' method is wrapped in this decorator.
        auth_ttl = parsed_args.authorization_ttl
        if auth_ttl is not None:
            if auth_ttl < 0:
                msg = _("%(param)s must be positive integer or zero.") % {
                    "param": "authorization-ttl"
                }
                raise exceptions.CommandError(msg)
            kwargs['authorization_ttl'] = auth_ttl

        federation_client.identity_providers.update(
            parsed_args.identity_provider, **kwargs
        )


class ShowIdentityProvider(command.ShowOne):
    _description = _("Display identity provider details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'identity_provider',
            metavar='<identity-provider>',
            help=_('Identity provider to display'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        idp = utils.find_resource(
            identity_client.federation.identity_providers,
            parsed_args.identity_provider,
            id=parsed_args.identity_provider,
        )

        idp._info.pop('links', None)
        remote_ids = format_columns.ListColumn(idp._info.pop('remote_ids', []))
        idp._info['remote_ids'] = remote_ids
        return zip(*sorted(idp._info.items()))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.cli import format_columns
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

_formatters = {
    'tags': format_columns.ListColumn,
}


def _format_task(task):
    """Format an task to make it more consistent with OSC operations."""

    info = {}
    properties = {}

    # the only fields we're not including is "links", "tags" and the properties
    fields_to_show = [
        'created_at',
        'expires_at',
        'id',
        'input',
        'message',
        'owner_id',
        'result',
        'status',
        'type',
        'updated_at',
    ]

    # split out the usual key and the properties which are top-level
    for field in fields_to_show:
        info[field] = task.get(field)

    for key in task:
        if key in fields_to_show:
            continue

        if key in {'location', 'name', 'schema'}:
            continue

        properties[key] = task.get(key)

    # add properties back into the dictionary as a top-level key
    info['properties'] = format_columns.DictColumn(properties)

    return info


class ShowTask(command.ShowOne):
    _description = _('Display task details')

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'task',
            metavar='<Task ID>',
            help=_('Task to display (ID)'),
        )

        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        task = image_client.get_task(parsed_args.task)
        info = _format_task(task)

        return zip(*sorted(info.items()))


class ListTask(command.Lister):
    _description = _('List tasks')

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--sort-key',
            metavar='<key>[:<field>]',
            help=_(
                'Sorts the response by one of the following attributes: '
                'created_at, expires_at, id, status, type, updated_at. '
                '(default is created_at) '
                '(multiple keys and directions can be specified separated '
                'by comma)'
            ),
        )
        parser.add_argument(
            '--sort-dir',
            metavar='<key>[:<direction>]',
            help=_(
                'Sort output by selected keys and directions (asc or desc) '
                '(default: name:desc) '
                '(multiple keys and directions can be specified separated '
                'by comma)'
            ),
        )
        parser.add_argument(
            '--limit',
            metavar='<num-tasks>',
            type=int,
            help=_('Maximum number of tasks to display.'),
        )
        parser.add_argument(
            '--marker',
            metavar='<task>',
            help=_(
                'The last task of the previous page. '
                'Display list of tasks after marker. '
                'Display all tasks if not specified. '
                '(name or ID)'
            ),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            choices=['import'],
            help=_('Filters the response by a task type.'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=[
                'pending',
                'processing',
                'success',
                'failure',
            ],
            help=_('Filter tasks based on status.'),
        )

        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        columns = ('id', 'type', 'status', 'owner_id')
        column_headers = ('ID', 'Type', 'Status', 'Owner')

        kwargs = {}
        copy_attrs = {
            'sort_key',
            'sort_dir',
            'limit',
            'marker',
            'type',
            'status',
        }
        for attr in copy_attrs:
            val = getattr(parsed_args, attr, None)
            if val is not None:
                # Only include a value in kwargs for attributes that are
                # actually present on the command line
                kwargs[attr] = val

        data = image_client.tasks(**kwargs)

        return (
            column_headers,
            (
                utils.get_item_properties(s, columns, formatters=_formatters)
                for s in data
            ),
        )



# Copyright 2023 Red Hat.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_image_cache(cached_images):
    """Format image cache to make it more consistent with OSC operations."""

    image_list = []
    for item in cached_images:
        if item == "cached_images":
            for image in cached_images[item]:
                image_obj = copy.deepcopy(image)
                image_obj['state'] = 'cached'
                image_obj['last_accessed'] = (
                    datetime.datetime.fromtimestamp(
                        image['last_accessed'], tz=datetime.timezone.utc
                    )
                    .replace(tzinfo=None)
                    .isoformat()
                )
                image_obj['last_modified'] = (
                    datetime.datetime.fromtimestamp(
                        image['last_modified'], tz=datetime.timezone.utc
                    )
                    .replace(tzinfo=None)
                    .isoformat()
                )
                image_list.append(image_obj)
        elif item == "queued_images":
            for image in cached_images[item]:
                image = {'image_id': image}
                image.update(
                    {
                        'state': 'queued',
                        'last_accessed': 'N/A',
                        'last_modified': 'N/A',
                        'size': 'N/A',
                        'hits': 'N/A',
                    }
                )
                image_list.append(image)
    return image_list


class ListCachedImage(command.Lister):
    _description = _("Get Cache State")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        # List of Cache data received
        data = _format_image_cache(dict(image_client.get_image_cache()))
        columns = [
            'image_id',
            'state',
            'last_accessed',
            'last_modified',
            'size',
            'hits',
        ]
        column_headers = [
            "ID",
            "State",
            "Last Accessed (UTC)",
            "Last Modified (UTC)",
            "Size",
            "Hits",
        ]

        return (
            column_headers,
            (
                utils.get_dict_properties(
                    image,
                    columns,
                )
                for image in data
            ),
        )


class QueueCachedImage(command.Command):
    _description = _("Queue image(s) for caching.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "images",
            metavar="<image>",
            nargs="+",
            help=_("Image to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        failures = 0
        for image in parsed_args.images:
            try:
                image_obj = image_client.find_image(
                    image,
                    ignore_missing=False,
                )
                image_client.queue_image(image_obj.id)
            except Exception as e:
                failures += 1
                msg = _(
                    "Failed to queue image with name or ID '%(image)s': %(e)s"
                )
                LOG.error(msg, {'image': image, 'e': e})

        if failures > 0:
            total = len(parsed_args.images)
            msg = _("Failed to queue %(failures)s of %(total)s images") % {
                'failures': failures,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class DeleteCachedImage(command.Command):
    _description = _("Delete image(s) from cache")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "images",
            metavar="<image>",
            nargs="+",
            help=_("Image(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        failures = 0
        image_client = self.app.client_manager.image
        for image in parsed_args.images:
            try:
                image_obj = image_client.find_image(
                    image,
                    ignore_missing=False,
                )
                image_client.cache_delete_image(image_obj.id)
            except Exception as e:
                failures += 1
                msg = _(
                    "Failed to delete image with name or ID '%(image)s': %(e)s"
                )
                LOG.error(msg, {'image': image, 'e': e})

        if failures > 0:
            total = len(parsed_args.images)
            msg = _("Failed to delete %(failures)s of %(total)s images.") % {
                'failures': failures,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ClearCachedImage(command.Command):
    _description = _("Clear all images from cache, queue or both")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--cache",
            action="store_const",
            const="cache",
            default="both",
            dest="target",
            help=_("Clears all the cached images"),
        )
        parser.add_argument(
            "--queue",
            action="store_const",
            const="queue",
            default="both",
            dest="target",
            help=_("Clears all the queued images"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        target = parsed_args.target
        try:
            image_client.clear_cache(target)
        except Exception:
            msg = _("Failed to clear image cache")
            LOG.error(msg)
            raise exceptions.CommandError(msg)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import abc
import contextlib
import logging
import typing as ty

from cliff import _argparse
import openstack.exceptions
from osc_lib.cli import parseractions
from osc_lib import exceptions

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.network import utils


LOG = logging.getLogger(__name__)

_required_opt_extensions_map = {
    'allowed_address_pairs': 'allowed-address-pairs',
    'dns_domain': 'dns-integration',
    'dns_name': 'dns-integration',
    'extra_dhcp_opts': 'extra_dhcp_opt',
    'qos_policy_id': 'qos',
    'security_groups': 'security-groups',
}

_NET_TYPE_NEUTRON = 'neutron'
_NET_TYPE_COMPUTE = 'nova-network'
_QUALIFIER_FMT = "%s\n\n*%s*"


@contextlib.contextmanager
def check_missing_extension_if_error(client_manager, attrs):
    # If specified option requires extension, then try to
    # find out if it exists. If it does not exist,
    # then an exception with the appropriate message
    # will be thrown from within client.find_extension
    try:
        yield
    except openstack.exceptions.HttpException:
        for opt, ext in _required_opt_extensions_map.items():
            if opt in attrs:
                client_manager.find_extension(ext, ignore_missing=False)
        raise


class NetDetectionMixin(metaclass=abc.ABCMeta):
    """Convenience methods for nova-network vs. neutron decisions.

    A live environment detects which network type it is running and creates its
    parser with only the options relevant to that network type.

    But the command classes are used for docs builds as well, and docs must
    present the options for both network types, often qualified accordingly.
    """

    @property
    def _network_type(self):
        """Discover whether the running cloud is using neutron or nova-network.

        :return:
            * ``NET_TYPE_NEUTRON`` if neutron is detected
            * ``NET_TYPE_COMPUTE`` if running in a cloud but neutron is not
              detected.
            * ``None`` if not running in a cloud, which hopefully means we're
              building docs.
        """
        # Have we set it up yet for this command?
        if not hasattr(self, '_net_type'):
            try:
                if self.app.client_manager.is_network_endpoint_enabled():  # type: ignore
                    net_type = _NET_TYPE_NEUTRON
                else:
                    net_type = _NET_TYPE_COMPUTE
            except AttributeError:
                LOG.warning(
                    "%s: Could not detect a network type. Assuming we are "
                    "building docs.",
                    self.__class__.__name__,
                )
                net_type = None
            self._net_type = net_type
        return self._net_type

    @property
    def is_neutron(self):
        return self._network_type is _NET_TYPE_NEUTRON

    @property
    def is_nova_network(self):
        return self._network_type is _NET_TYPE_COMPUTE

    @property
    def is_docs_build(self):
        return self._network_type is None

    def enhance_help_neutron(self, _help):
        if self.is_docs_build:
            # Why can't we say 'neutron'?
            return _QUALIFIER_FMT % (_help, _("Network version 2 only"))
        return _help

    def enhance_help_nova_network(self, _help):
        if self.is_docs_build:
            # Why can't we say 'nova-network'?
            return _QUALIFIER_FMT % (_help, _("Compute version 2 only"))
        return _help

    @staticmethod
    def split_help(network_help, compute_help):
        return (
            "*{network_qualifier}:*\n  {network_help}\n\n"
            "*{compute_qualifier}:*\n  {compute_help}".format(
                **dict(
                    network_qualifier=_("Network version 2"),
                    network_help=network_help,
                    compute_qualifier=_("Compute version 2"),
                    compute_help=compute_help,
                )
            )
        )

    def get_parser(self, prog_name: str) -> _argparse.ArgumentParser:
        LOG.debug('get_parser(%s)', prog_name)
        parser = super().get_parser(prog_name)  # type: ignore
        parser = self.update_parser_common(parser)
        LOG.debug('common parser: %s', parser)
        if self.is_neutron or self.is_docs_build:
            parser = self.update_parser_network(parser)
        if self.is_nova_network or self.is_docs_build:
            # Add nova-net options if running nova-network or building docs
            parser = self.update_parser_compute(parser)
        return parser

    def update_parser_common(self, parser):
        """Default is no updates to parser."""
        return parser

    def update_parser_network(self, parser):
        """Default is no updates to parser."""
        return parser

    def update_parser_compute(self, parser):
        """Default is no updates to parser."""
        return parser

    def take_action(self, parsed_args):
        if self.is_neutron:
            return self.take_action_network(
                self.app.client_manager.network,  # type: ignore
                parsed_args,
            )
        elif self.is_nova_network:
            return self.take_action_compute(
                self.app.client_manager.compute,  # type: ignore
                parsed_args,
            )

    def take_action_network(self, client, parsed_args):
        """Override to do something useful."""
        pass

    def take_action_compute(self, client, parsed_args):
        """Override to do something useful."""
        pass


class NetworkAndComputeCommand(
    NetDetectionMixin, command.Command, metaclass=abc.ABCMeta
):
    """Network and Compute Command

    Command class for commands that support implementation via
    the network or compute endpoint. Such commands have different
    implementations for take_action() and may even have different
    arguments.
    """

    pass


class NetworkAndComputeDelete(NetworkAndComputeCommand, metaclass=abc.ABCMeta):
    """Network and Compute Delete

    Delete class for commands that support implementation via
    the network or compute endpoint. Such commands have different
    implementations for take_action() and may even have different
    arguments. This class supports bulk deletion, and error handling
    following the rules in doc/source/command-errors.rst.
    """

    resource: str

    def take_action(self, parsed_args):
        ret = 0
        resources = getattr(parsed_args, self.resource, [])

        for r in resources:
            self.r = r
            try:
                if self.app.client_manager.is_network_endpoint_enabled():
                    self.take_action_network(
                        self.app.client_manager.network, parsed_args
                    )
                else:
                    self.take_action_compute(
                        self.app.client_manager.compute,
                        parsed_args,
                    )
            except Exception as e:
                msg = _(
                    "Failed to delete %(resource)s with name or ID "
                    "'%(name_or_id)s': %(e)s"
                ) % {
                    "resource": self.resource,
                    "name_or_id": r,
                    "e": e,
                }
                LOG.error(msg)
                ret += 1

        if ret:
            total = len(resources)
            msg = _("%(num)s of %(total)s %(resource)ss failed to delete.") % {
                "num": ret,
                "total": total,
                "resource": self.resource,
            }
            raise exceptions.CommandError(msg)


class NetworkAndComputeLister(
    NetDetectionMixin, command.Lister, metaclass=abc.ABCMeta
):
    """Network and Compute Lister

    Lister class for commands that support implementation via
    the network or compute endpoint. Such commands have different
    implementations for take_action() and may even have different
    arguments.
    """

    pass


class NetworkAndComputeShowOne(
    NetDetectionMixin, command.ShowOne, metaclass=abc.ABCMeta
):
    """Network and Compute ShowOne

    ShowOne class for commands that support implementation via
    the network or compute endpoint. Such commands have different
    implementations for take_action() and may even have different
    arguments.
    """

    def take_action(self, parsed_args):
        try:
            if self.app.client_manager.is_network_endpoint_enabled():
                return self.take_action_network(
                    self.app.client_manager.network, parsed_args
                )
            else:
                return self.take_action_compute(
                    self.app.client_manager.compute, parsed_args
                )
        except openstack.exceptions.HttpException as exc:
            msg = _("Error while executing command: %s") % exc.message
            if exc.details:
                msg += ", " + str(exc.details)
            raise exceptions.CommandError(msg)


class NeutronCommandWithExtraArgs(command.Command):
    """Create and Update commands with additional extra properties.

    Extra properties can be passed to the command and are then send to the
    Neutron as given to the command.
    """

    # dict of allowed types
    _allowed_types_dict = {
        'bool': utils.str2bool,
        'dict': utils.str2dict,
        'list': utils.str2list,
        'int': int,
        'str': str,
    }

    def _get_property_converter(self, _property):
        if 'type' in _property:
            converter = self._allowed_types_dict.get(_property['type'])
        else:
            converter = str

        if not converter:
            raise exceptions.CommandError(
                _(
                    "Type {property_type} of property {name} is not supported"
                ).format(
                    property_type=_property['type'], name=_property['name']
                )
            )
        return converter

    def _parse_extra_properties(self, extra_properties):
        result: dict[str, ty.Any] = {}
        if extra_properties:
            for _property in extra_properties:
                converter = self._get_property_converter(_property)
                result[_property['name']] = converter(_property['value'])
        return result

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--extra-property',
            metavar='type=<property_type>,name=<property_name>,'
            'value=<property_value>',
            dest='extra_properties',
            action=parseractions.MultiKeyValueAction,
            required_keys=['name', 'value'],
            optional_keys=['type'],
            help=_(
                "Additional parameters can be passed using this property. "
                "Default type of the extra property is string ('str'), but "
                "other types can be used as well. Available types are: "
                "'dict', 'list', 'str', 'bool', 'int'. "
                "In case of 'list' type, 'value' can be "
                "semicolon-separated list of values. "
                "For 'dict' value is semicolon-separated list of the "
                "key:value pairs."
            ),
        )
        return parser


class NeutronUnsetCommandWithExtraArgs(NeutronCommandWithExtraArgs):
    def _parse_extra_properties(self, extra_properties):
        result: dict[str, ty.Any] = {}
        if extra_properties:
            for _property in extra_properties:
                result[_property['name']] = None

        return result



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


from openstack.identity.v3 import region as _region
from openstack.test import fakes as sdk_fakes

from openstackclient.identity.v3 import region
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRegionCreate(identity_fakes.TestIdentityv3):
    region = sdk_fakes.generate_fake_resource(_region.Region)
    columns = (
        'region',
        'description',
        'parent_region',
    )
    datalist = (
        region.id,
        region.description,
        region.parent_region_id,
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.create_region.return_value = self.region

        # Get the command object to test
        self.cmd = region.CreateRegion(self.app, None)

    def test_region_create_description(self):
        arglist = [
            self.region.id,
            '--description',
            self.region.description,
        ]
        verifylist = [
            ('region', self.region.id),
            ('description', self.region.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.region.description,
            'id': self.region.id,
            'parent_region_id': None,
        }
        self.identity_sdk_client.create_region.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_region_create_no_options(self):
        arglist = [
            self.region.id,
        ]
        verifylist = [
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'id': self.region.id,
            'parent_region_id': None,
        }
        self.identity_sdk_client.create_region.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_region_create_parent_region_id(self):
        arglist = [
            self.region.id,
            '--parent-region',
            self.region.parent_region_id,
        ]
        verifylist = [
            ('region', self.region.id),
            ('parent_region', self.region.parent_region_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': None,
            'id': self.region.id,
            'parent_region_id': self.region.parent_region_id,
        }
        self.identity_sdk_client.create_region.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestRegionDelete(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.region = sdk_fakes.generate_fake_resource(_region.Region)
        self.identity_sdk_client.delete_region.return_value = None

        # Get the command object to test
        self.cmd = region.DeleteRegion(self.app, None)

    def test_region_delete_no_options(self):
        arglist = [
            self.region.id,
        ]
        verifylist = [
            ('region', [self.region.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_region.assert_called_with(
            self.region.id,
        )
        self.assertIsNone(result)


class TestRegionList(identity_fakes.TestIdentityv3):
    region = sdk_fakes.generate_fake_resource(_region.Region)
    columns = (
        'Region',
        'Parent Region',
        'Description',
    )
    datalist = (
        (
            region.id,
            region.parent_region_id,
            region.description,
        ),
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.regions.return_value = [self.region]

        # Get the command object to test
        self.cmd = region.ListRegion(self.app, None)

    def test_region_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.regions.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_region_list_parent_region_id(self):
        arglist = [
            '--parent-region',
            self.region.parent_region_id,
        ]
        verifylist = [
            ('parent_region', self.region.parent_region_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.regions.assert_called_with(
            parent_region_id=self.region.parent_region_id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))


class TestRegionSet(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.region = sdk_fakes.generate_fake_resource(_region.Region)

        # Get the command object to test
        self.cmd = region.SetRegion(self.app, None)

    def test_region_set_no_options(self):
        arglist = [
            self.region.id,
        ]
        verifylist = [
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {}
        self.identity_sdk_client.update_region.assert_called_with(
            self.region.id, **kwargs
        )
        self.assertIsNone(result)

    def test_region_set_description(self):
        arglist = [
            '--description',
            'qwerty',
            self.region.id,
        ]
        verifylist = [
            ('description', 'qwerty'),
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': 'qwerty',
        }
        self.identity_sdk_client.update_region.assert_called_with(
            self.region.id, **kwargs
        )
        self.assertIsNone(result)

    def test_region_set_parent_region_id(self):
        arglist = [
            '--parent-region',
            'new_parent',
            self.region.id,
        ]
        verifylist = [
            ('parent_region', 'new_parent'),
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'parent_region_id': 'new_parent',
        }
        self.identity_sdk_client.update_region.assert_called_with(
            self.region.id, **kwargs
        )
        self.assertIsNone(result)


class TestRegionShow(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.region = sdk_fakes.generate_fake_resource(_region.Region)
        self.identity_sdk_client.get_region.return_value = self.region

        # Get the command object to test
        self.cmd = region.ShowRegion(self.app, None)

    def test_region_show(self):
        arglist = [
            self.region.id,
        ]
        verifylist = [
            ('region', self.region.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.get_region.assert_called_with(
            self.region.id,
        )

        collist = ('region', 'description', 'parent_region')
        self.assertEqual(collist, columns)
        datalist = (
            self.region.id,
            self.region.description,
            self.region.parent_region_id,
        )
        self.assertEqual(datalist, data)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib import utils

from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

# global variables used when building the shell
DEFAULT_API_VERSION = '2'
API_VERSION_OPTION = 'os_image_api_version'
API_NAME = 'image'
API_VERSIONS = ('1', '2')


def make_client(instance):
    """Returns an image service client."""
    LOG.debug(
        'Image client initialized using OpenStack SDK: %s',
        instance.sdk_connection.image,
    )
    return instance.sdk_connection.image


def build_option_parser(parser):
    """Hook to add global options"""
    parser.add_argument(
        '--os-image-api-version',
        metavar='<image-api-version>',
        default=utils.env('OS_IMAGE_API_VERSION'),
        help=_('Image API version, default=%s (Env: OS_IMAGE_API_VERSION)')
        % DEFAULT_API_VERSION,
    )
    return parser


def check_api_version(check_version):
    # SDK supports auto-negotiation for us: always return True
    return True



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib import exceptions
from osc_lib import utils

from openstackclient.tests.unit import utils as test_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_transfer_request


class TestTransfer(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the TransferManager Mock
        self.transfer_mock = self.volume_client.transfers
        self.transfer_mock.reset_mock()

        # Get a shortcut to the VolumeManager Mock
        self.volumes_mock = self.volume_client.volumes
        self.volumes_mock.reset_mock()


class TestTransferAccept(TestTransfer):
    columns = (
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume_transfer = volume_fakes.create_one_transfer()
        self.data = (
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.transfer_mock.get.return_value = self.volume_transfer
        self.transfer_mock.accept.return_value = self.volume_transfer

        # Get the command object to test
        self.cmd = volume_transfer_request.AcceptTransferRequest(
            self.app, None
        )

    def test_transfer_accept(self):
        arglist = [
            '--auth-key',
            'key_value',
            self.volume_transfer.id,
        ]
        verifylist = [
            ('transfer_request', self.volume_transfer.id),
            ('auth_key', 'key_value'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfer_mock.get.assert_called_once_with(
            self.volume_transfer.id,
        )
        self.transfer_mock.accept.assert_called_once_with(
            self.volume_transfer.id,
            'key_value',
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_accept_no_option(self):
        arglist = [
            self.volume_transfer.id,
        ]
        verifylist = [
            ('transfer_request', self.volume_transfer.id),
        ]

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestTransferCreate(TestTransfer):
    volume = volume_fakes.create_one_volume()

    columns = (
        'auth_key',
        'created_at',
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume_transfer = volume_fakes.create_one_transfer(
            attrs={
                'volume_id': self.volume.id,
                'auth_key': 'key',
                'created_at': 'time',
            },
        )
        self.data = (
            self.volume_transfer.auth_key,
            self.volume_transfer.created_at,
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.transfer_mock.create.return_value = self.volume_transfer
        self.volumes_mock.get.return_value = self.volume

        # Get the command object to test
        self.cmd = volume_transfer_request.CreateTransferRequest(
            self.app, None
        )

    def test_transfer_create_without_name(self):
        arglist = [
            self.volume.id,
        ]
        verifylist = [
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfer_mock.create.assert_called_once_with(self.volume.id, None)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_with_name(self):
        arglist = [
            '--name',
            self.volume_transfer.name,
            self.volume.id,
        ]
        verifylist = [
            ('name', self.volume_transfer.name),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfer_mock.create.assert_called_once_with(
            self.volume.id,
            self.volume_transfer.name,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_with_no_snapshots(self):
        self.set_volume_api_version('3.55')

        arglist = [
            '--no-snapshots',
            self.volume.id,
        ]
        verifylist = [
            ('name', None),
            ('snapshots', False),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfer_mock.create.assert_called_once_with(
            self.volume.id, None, no_snapshots=True
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_transfer_create_pre_v355(self):
        self.set_volume_api_version('3.54')

        arglist = [
            '--no-snapshots',
            self.volume.id,
        ]
        verifylist = [
            ('name', None),
            ('snapshots', False),
            ('volume', self.volume.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.55 or greater is required', str(exc)
        )


class TestTransferDelete(TestTransfer):
    volume_transfers = volume_fakes.create_transfers(count=2)

    def setUp(self):
        super().setUp()

        self.transfer_mock.get = volume_fakes.get_transfers(
            self.volume_transfers,
        )
        self.transfer_mock.delete.return_value = None

        # Get the command object to mock
        self.cmd = volume_transfer_request.DeleteTransferRequest(
            self.app, None
        )

    def test_transfer_delete(self):
        arglist = [self.volume_transfers[0].id]
        verifylist = [("transfer_request", [self.volume_transfers[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.transfer_mock.delete.assert_called_with(
            self.volume_transfers[0].id
        )
        self.assertIsNone(result)

    def test_delete_multiple_transfers(self):
        arglist = []
        for v in self.volume_transfers:
            arglist.append(v.id)
        verifylist = [
            ('transfer_request', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for v in self.volume_transfers:
            calls.append(call(v.id))
        self.transfer_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_transfers_with_exception(self):
        arglist = [
            self.volume_transfers[0].id,
            'unexist_transfer',
        ]
        verifylist = [
            ('transfer_request', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self.volume_transfers[0], exceptions.CommandError]
        with mock.patch.object(
            utils, 'find_resource', side_effect=find_mock_result
        ) as find_mock:
            try:
                self.cmd.take_action(parsed_args)
                self.fail('CommandError should be raised.')
            except exceptions.CommandError as e:
                self.assertEqual(
                    '1 of 2 volume transfer requests failed to delete',
                    str(e),
                )

            find_mock.assert_any_call(
                self.transfer_mock, self.volume_transfers[0].id
            )
            find_mock.assert_any_call(self.transfer_mock, 'unexist_transfer')

            self.assertEqual(2, find_mock.call_count)
            self.transfer_mock.delete.assert_called_once_with(
                self.volume_transfers[0].id,
            )


class TestTransferList(TestTransfer):
    # The Transfers to be listed
    volume_transfers = volume_fakes.create_one_transfer()

    def setUp(self):
        super().setUp()

        self.transfer_mock.list.return_value = [self.volume_transfers]

        # Get the command object to test
        self.cmd = volume_transfer_request.ListTransferRequest(self.app, None)

    def test_transfer_list_without_argument(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'ID',
            'Name',
            'Volume',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = (
            (
                self.volume_transfers.id,
                self.volume_transfers.name,
                self.volume_transfers.volume_id,
            ),
        )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to list volume_transfers
        self.transfer_mock.list.assert_called_with(
            detailed=True, search_opts={'all_tenants': 0}
        )

    def test_transfer_list_with_argument(self):
        arglist = ["--all-projects"]
        verifylist = [("all_projects", True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'ID',
            'Name',
            'Volume',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = (
            (
                self.volume_transfers.id,
                self.volume_transfers.name,
                self.volume_transfers.volume_id,
            ),
        )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to list volume_transfers
        self.transfer_mock.list.assert_called_with(
            detailed=True, search_opts={'all_tenants': 1}
        )


class TestTransferShow(TestTransfer):
    columns = (
        'created_at',
        'id',
        'name',
        'volume_id',
    )

    def setUp(self):
        super().setUp()

        self.volume_transfer = volume_fakes.create_one_transfer(
            attrs={'created_at': 'time'},
        )
        self.data = (
            self.volume_transfer.created_at,
            self.volume_transfer.id,
            self.volume_transfer.name,
            self.volume_transfer.volume_id,
        )

        self.transfer_mock.get.return_value = self.volume_transfer

        # Get the command object to test
        self.cmd = volume_transfer_request.ShowTransferRequest(self.app, None)

    def test_transfer_show(self):
        arglist = [
            self.volume_transfer.id,
        ]
        verifylist = [
            ('transfer_request', self.volume_transfer.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfer_mock.get.assert_called_once_with(self.volume_transfer.id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstackclient.tests.functional import base


class ModuleTest(base.TestCase):
    """Functional tests for openstackclient module list output."""

    CLIENTS = ['openstackclient', 'keystoneclient', 'openstack']

    LIBS = ['osc_lib', 'keystoneauth1']

    def test_module_list(self):
        # Test module list
        cmd_output = self.openstack('module list', parse_output=True)
        for one_module in self.CLIENTS:
            self.assertIn(one_module, cmd_output.keys())
        for one_module in self.LIBS:
            self.assertNotIn(one_module, cmd_output.keys())

        # Test module list --all
        cmd_output = self.openstack('module list --all', parse_output=True)
        for one_module in self.CLIENTS + self.LIBS:
            self.assertIn(one_module, cmd_output.keys())


class CommandTest(base.TestCase):
    """Functional tests for openstackclient command list."""

    GROUPS = [
        'openstack.volume.v3',
        'openstack.network.v2',
        'openstack.image.v2',
        'openstack.identity.v3',
        'openstack.compute.v2',
        'openstack.common',
        'openstack.cli',
    ]

    def test_command_list_no_option(self):
        cmd_output = self.openstack('command list', parse_output=True)
        group_names = [each.get('Command Group') for each in cmd_output]
        for one_group in self.GROUPS:
            self.assertIn(one_group, group_names)

    def test_command_list_with_group(self):
        input_groups = ['volume', 'network', 'image', 'identity', 'compute.v2']
        for each_input in input_groups:
            cmd_output = self.openstack(
                f'command list --group {each_input}',
                parse_output=True,
            )
            group_names = [each.get('Command Group') for each in cmd_output]
            for each_name in group_names:
                self.assertIn(each_input, each_name)
